/*
   arguments.c
   
   THIS SOFTWARE IS IN PUBLIC DOMAIN
     There is no limitation on any kind of usage of this code. 
     You use it because it is yours.
     
     Originally written and maintained by Matlib (matlib@matlibhax.com),
     any comments or suggestions are welcome.
     
   Change log:
        1  2010-02-24  Matlib
           2011-02-16  update code to use libarguments v. 3
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>

#include <arguments.h>


#define CHECK_ERROR(__COND__, __FUN__) \
  if (__COND__) \
  { \
    perror (__FUN__); \
    exit (EXIT_FAILURE); \
  }


struct passwd *defaults;



static uid_t get_uid (char const *const name)
{
  char *tail;
  long int uid;

  uid = strtol (name, &tail, 0);
  if ((*tail == 0) && (uid >= 0))
    return uid;
  if ((defaults = getpwnam (name)) == NULL)
    arguments_error_freeform ("User \"%s\" not found in password database.", 
       name);
  return defaults -> pw_uid;
}

static gid_t get_gid (char const *const name)
{
  char *tail;
  long int gid;
  struct group *g;

  gid = strtol (name, &tail, 0);
  if ((*tail == 0) && (gid >= 0))
    return gid;
  if ((g = getgrnam (name)) == NULL)
    arguments_error_freeform ("Group \"%s\" not found in group database.", 
       name);
  return g -> gr_gid;
}

static void get_defaults ()
{
  if (defaults != NULL)
    return;
  defaults = getpwuid (getuid ());
  if (defaults == NULL)
  {
    fprintf (stderr, "Cannot retrieve user settings for UID %d.", getuid ());
    exit (EXIT_FAILURE);
  }
}




static ARGUMENT_CALLBACK (argument_uid)
{
  arg -> value_int = get_uid (parameter);
}

static ARGUMENT_CALLBACK (argument_gid)
{
  arg -> value_int = get_gid (parameter);
}

gid_t *groups;
int gp, gl;
#define GROUP_BLOCK 4

static ARGUMENT_CALLBACK (argument_groups)
{
  if (gp == gl)
  {
    gl += GROUP_BLOCK;
    groups = realloc (groups, sizeof (gid_t) * gl);
    CHECK_ERROR (groups == NULL, "realloc");
  }
  groups [gp] = get_gid (parameter);
  gp ++;
}



struct Argument arguments [] =
{
  {'u', "ruser",        argument_uid,    ARGUMENT_F_ONCE | ARGUMENT_F_PAR, {.value_int = -1}},
  {'U', "euser",        argument_uid,    ARGUMENT_F_ONCE | ARGUMENT_F_PAR, {.value_int = -1}},
  {'g', "rgroup",       argument_gid,    ARGUMENT_F_ONCE | ARGUMENT_F_PAR, {.value_int = -1}},
  {'G', "egroup",       argument_gid,    ARGUMENT_F_ONCE | ARGUMENT_F_PAR, {.value_int = -1}},
  {'c', "clear-groups", argument_bool,   ARGUMENT_F_ONCE, {.value = NULL}},
  {'s', "add-group",    argument_groups, ARGUMENT_F_PAR, {.value = NULL}},
  {'n', "program-name", argument_vs,     ARGUMENT_F_ONCE | ARGUMENT_F_PAR, {.value = NULL}},
  {}
};

enum
{
  ARG_N_RUID,
  ARG_N_EUID,
  ARG_N_RGID,
  ARG_N_EGID,
  ARG_N_CLRG,
  ARG_N_ADDG,
  ARG_N_NAME
};

#define ARG_RUID (arguments [ARG_N_RUID].value_int)
#define ARG_EUID (arguments [ARG_N_EUID].value_int)
#define ARG_RGID (arguments [ARG_N_RGID].value_int)
#define ARG_EGID (arguments [ARG_N_EGID].value_int)
#define ARG_CLRG (arguments [ARG_N_CLRG].value_u)
#define ARG_NAME (arguments [ARG_N_NAME].value_vs)




static void set_defaults (char const *const name)
{
  ARG_RUID = ARG_EUID = get_uid (name);
  get_defaults ();
  ARG_RGID = ARG_EGID = defaults -> pw_gid;
  ARG_CLRG = 1;
}


static void set_groups ()
{
  gid_t *g;
  int n, i, j;
  
  if (ARG_CLRG)
  {
    CHECK_ERROR (setgroups (gp, groups) != 0, "setgroups");
    return;
  }
  n = getgroups (0, NULL);
  CHECK_ERROR (n == -1, "getgroups: get number of groups");
  if (n == 0)
  {
    CHECK_ERROR (setgroups (gp, groups) != 0, "setgroups");
    return;
  }
  g = malloc (sizeof (gid_t) * (n + gp));
  i = getgroups (n, g);
  CHECK_ERROR (i == -1, "getgroups");
  if (i != n)
  {
    fprintf (stderr, "Calls to getgroups returned different number of "
       "supplementary groups. Confused, aborting.\n");
    exit (EXIT_FAILURE);
  }
  
  i = 0;
  while (i < gp)
  {
    for (j = 0; j < n; j ++)
      if (g [j] == groups [i])
        goto lbreak;
    g [n ++] = groups [i];
lbreak:
    i ++;
  }
  CHECK_ERROR (setgroups (n, g) != 0, "setgroups");
}


int main (int const argc, char *const *const argv)
{
  int n;
  char **a;
  char *progname;
  
  defaults = NULL;
  gp = 0;
  gl = GROUP_BLOCK;
  groups = malloc (sizeof (gid_t) * GROUP_BLOCK);
  CHECK_ERROR (groups == NULL, "malloc");
  
  n = arguments_parse ("setuidgid", "1 (2010-02-24)", "[command ...]", 
     "Set real or effective process IDs and execute a command.",
     argc, (void *) argv, arguments);
     
  if ((ARG_RUID == -1) && (ARG_RGID == -1) && (ARG_EUID == -1)
     && (ARG_RGID == -1))
  {
    if (n == argc)
      arguments_error_freeform ("User name required.");
    set_defaults (argv [n]);
    n ++;
  }
  
  if (ARG_CLRG || (gp != 0))
    set_groups ();
  CHECK_ERROR (setregid (ARG_RGID, ARG_EGID) != 0, "setregid");
  CHECK_ERROR (setreuid (ARG_RUID, ARG_EUID) != 0, "setreuid");
  
  if (n == argc)
  {
    get_defaults ();
    progname = defaults -> pw_shell;
    if (progname == NULL)
    {
      fprintf (stderr, 
         "User's default shell is not specified in password database.\n");
      exit (EXIT_FAILURE);
    }
  }
  else
    progname = argv [n ++];
  a = malloc (sizeof (char *) * (argc - n + 2));
  a [0] = ARG_NAME == NULL ? progname : ARG_NAME;
  for (gp = 1; n < argc; gp ++)
    a [gp] = argv [n ++];
  a [gp] = NULL;
  execvp (progname, a);
  perror (progname);
  return EXIT_FAILURE;
}

